/***************************************************************************
* Additional implementation of "BIKE: Bit Flipping Key Encapsulation". 
* Copyright (c) 2017 Nir Drucker, Shay Gueron
* (drucker.nir@gmail.com, shay.gueron@gmail.com)
* The license is detailed in the file LICENSE.txt, and applies to this file.
* ***************************************************************************/

#ifndef __BIKE_DEFS_H_INCLUDED__
#define __BIKE_DEFS_H_INCLUDED__

#include "defs.h"

////////////////////////////////////////////
//             BIKE Parameters
///////////////////////////////////////////
#define N0 2

#ifndef LEVEL
  #define LEVEL 5
#endif

#if LEVEL==5
    // 128-bits of post-quantum security parameters (BIKE paper):
    // Increased r for better DFR
    #define R_BITS  32749
    #define DV      137
    #define FAKE_DV 261
    #define T1      264
    #define FIXED_TH
    
    //The gfm code is optimized to a block size in this case:
    #define BLOCK_SIZE (32768)
#else
#if LEVEL==3
    #define R_BITS  21893
    #define DV      103
    #define FAKE_DV 131
    #define T1      199
    #define ADHOC_TH

    //The gfm code is optimized to a block size in this case:
    #define BLOCK_SIZE (32768)
#else
#if LEVEL==1
    // 64-bits of post-quantum security parameters (BIKE paper):
    #define R_BITS  10163
    #define DV      71
    #define FAKE_DV 90
    #define T1      134
    #define ADHOC_TH

    //The gfm code is optimized to a block size in this case:
    #define BLOCK_SIZE (16384)
#else
    #error "Bad level, choose one of 1/3/5"
#endif
#endif
#endif

#define MAX_DELTA 4

//Round the size to the nearest byte.
//SIZE suffix, is the number of bytes (uint8_t).
#define N_BITS   (R_BITS * N0)
#define R_SIZE   DIVIDE_AND_CEIL(R_BITS, 8)
#define R_QW     DIVIDE_AND_CEIL(R_BITS, 64)
#define N_SIZE   DIVIDE_AND_CEIL(N_BITS, 8)
#define N_QW     DIVIDE_AND_CEIL(N_BITS, 64)
#define N_EXTRA_BYTES (8*N_QW - N_SIZE)

#define R_BLOCKS         DIVIDE_AND_CEIL(R_BITS,BLOCK_SIZE)
#define R_PADDED        (R_BLOCKS * BLOCK_SIZE)
#define R_PADDED_SIZE   (R_PADDED/8)
#define R_PADDED_QW     (R_PADDED/64)

#define N_BLOCKS         DIVIDE_AND_CEIL(N_BITS,BLOCK_SIZE)
#define N_PADDED        (N_BLOCKS * BLOCK_SIZE)
#define N_PADDED_SIZE   (N_PADDED/8)
#define N_PADDED_QW     (N_PADDED/64)

#define R_DQWORDS DIVIDE_AND_CEIL(R_SIZE, 16)

#ifdef AVX512
#define R_QDQWORDS_BITS (DIVIDE_AND_CEIL(R_BITS, ALL_ZMM_SIZE) * ALL_ZMM_SIZE)
static_assert((R_BITS % ALL_ZMM_SIZE != 0), rbits_2048_err);

#define N_QDQWORDS_BITS (DIVIDE_AND_CEIL(N_BITS, ALL_ZMM_SIZE) * ALL_ZMM_SIZE)
static_assert((N_BITS % ALL_ZMM_SIZE!= 0), nbits_2048_err);

#else

#define R_DDQWORDS_BITS (DIVIDE_AND_CEIL(R_BITS, ALL_YMM_SIZE) * ALL_YMM_SIZE)
static_assert((R_BITS % ALL_YMM_SIZE != 0), rbits_512_err);

#define N_DDQWORDS_BITS (DIVIDE_AND_CEIL(N_BITS, ALL_YMM_SIZE) * ALL_YMM_SIZE)
static_assert((N_BITS % ALL_YMM_SIZE != 0), nbits_512_err);

#endif

#define LAST_R_QW_LEAD  (R_BITS & MASK(6))
#define LAST_R_QW_TRAIL (64 - LAST_R_QW_LEAD)
#define LAST_R_QW_MASK  MASK(LAST_R_QW_LEAD)

#define LAST_R_BYTE_LEAD  (R_BITS & MASK(3))
#define LAST_R_BYTE_TRAIL (8 - LAST_R_BYTE_LEAD)
#define LAST_R_BYTE_MASK  MASK(LAST_R_BYTE_LEAD)

// BIKE auxiliary functions parameters:
#define ELL_K_BITS  256
#define ELL_K_SIZE (ELL_K_BITS/8)

#endif //__BIKE_DEFS_H_INCLUDED__

